#include <stdlib.h>
#include <time.h>
#include "x_alloc.hpp"
#include "x_err.hpp"
#include "unique.hpp"
#include "packet.hpp"
#include "port.hpp"

#ifdef DEBUG
#include <stdio.h>
#endif

#ifdef IPXEMU
#include "ipx.hpp"
#endif

int			UARTtype;
int                     fifoSafety = 1;
int			port;
int			base;
int			irq;
dword			bps;

static void interrupt	(*old_commvect)();
static Queue		inQ, outQ;
static int		lineStatus;
static int		emptyTrans;
static int		commvectHooked;
static int		commvectNum;
#ifdef DEBUG
static time_t           firstTime, lastTime;
static dword            nSent;
static dword		nISRcalls;
static dword            maxIcount;
static dword            maxOcount;
static dword            nIdropped;
static dword            nOdropped;
#endif

void			portInit();
void			portDeinit();
int			portRead(byte *c);
int			portWrite(int);
int			portWrite(byte *s, int l);
int			portIcount();
int			portOcount();
void			portIclear();
void __interrupt	ISR_8250();
void __interrupt	ISR_16550();

void portInit() {
  byte mcr;
  dword divisor;
  byte temp;
  emptyTrans = 1;
  inQ.data = (byte*)smalloc(I_QSIZE * sizeof(byte));
  outQ.data = (byte*)smalloc(O_QSIZE * sizeof(byte));
  divisor = CLOCK_FREQUENCY / (16 * bps);
  OUT(base + LINE_CONTROL_REGISTER, LCR_DLAB);
  OUT(base + DIVISOR_LATCH_HIGH, 0);
  OUT(base + DIVISOR_LATCH_LOW, (byte) divisor);
  OUT(base + LINE_CONTROL_REGISTER, 3);
  /* check for a 16550 */
  UARTtype = UART_8250;
  if (fifoSafety != 4) {
    OUT(base + FIFO_CONTROL_REGISTER, FCR_FIFO_ENABLE + FCR_TRIGGER_14 - fifoSafety * 0x40);
    temp = IN(base + INTERRUPT_ID_REGISTER);
    if ((temp & 0xf8) == 0xc0) UARTtype = UART_16550;
  }
  if (UARTtype == UART_8250)
    OUT(base + FIFO_CONTROL_REGISTER, 0);
  /* prepare for interrupts */
  OUT(base + INTERRUPT_ENABLE_REGISTER, 0);
  mcr = IN(base + MODEM_CONTROL_REGISTER);
  mcr |= MCR_OUT2;
  mcr &= ~MCR_LOOPBACK;
  mcr |= MCR_DTR;
  mcr |= MCR_RTS;
  OUT(base + MODEM_CONTROL_REGISTER, mcr);
  IN(base);
  IN(base + INTERRUPT_ID_REGISTER);
  commvectNum = irq + 8;
  if (!commvectHooked)	{
    old_commvect = _dos_getvect(commvectNum);
    commvectHooked = 1;
    if (UARTtype == UART_8250) _dos_setvect(commvectNum, ISR_8250);
    else _dos_setvect(commvectNum, ISR_16550);
  }
  OUT(0x20 + 1, IN(0x20 + 1) & ~(1 << irq));
  cli();
  OUT(base + INTERRUPT_ENABLE_REGISTER,
      IER_RX_DATA_READY + IER_TX_HOLDING_REGISTER_EMPTY + IER_LINE_STATUS);
  OUT(0x20, 0xc2);
  sti();
}

void portDeinit() {
#ifdef DEBUG
  FILE *f = fopen("debug.txt", "at");
  fprintf(f, "--- PORT Module Debug Info -------------------------------------\n");
  fprintf(f, "nISRcalls:  %lu\n", nISRcalls);
  fprintf(f, "maxIcount:  %lu\n", maxIcount);
  fprintf(f, "maxOcount:  %lu\n", maxOcount);
  fprintf(f, "nIdropped:  %lu\n", nIdropped);
  fprintf(f, "nOdropped:  %lu\n", nOdropped);
  fprintf(f, "data rate:  %lu\n", nSent / (lastTime - firstTime));
  fclose(f);
#endif
  sti();
  while (portOcount());
  delay(250);
  if (!commvectHooked) return;
  OUT(base + INTERRUPT_ENABLE_REGISTER, 0);
  OUT(base + MODEM_CONTROL_REGISTER, MCR_DTR);
  OUT(base + FIFO_CONTROL_REGISTER, 0);
  OUT(0x20 + 1, IN(0x20 + 1) | (1 << irq));
  cli();
  _dos_setvect(commvectNum, old_commvect);
  sti();
  free(inQ.data);
  free(outQ.data);
}

int portRead(byte *c) {
  if (!inQ.n) return 1;
  *c = inQ.data[inQ.tail];
  if (++inQ.tail == I_QSIZE) inQ.tail = 0;
  inQ.n--;
  return 0;
}

int portWrite(int c) {
  if (outQ.n == O_QSIZE) {
#ifdef DEBUG
    nOdropped++;
#endif
    return 1;
  }
#ifdef DEBUG
  if (!firstTime) firstTime = time(NULL);
  lastTime = time(NULL);
  nSent++;
#endif
  if (emptyTrans) {
    emptyTrans = 0;
    OUT(base + TRANSMIT_HOLDING_REGISTER, c);
  } else {
    outQ.data[outQ.head] = c;
    if (++outQ.head == O_QSIZE) outQ.head = 0;
    outQ.n++;
  }
  return 0;
}

int portWrite(byte *s, int l) {
  if (l > O_QSIZE - outQ.n) {
#ifdef DEBUG
    nOdropped += l;
#endif
    return 1;
  }
#ifdef DEBUG
  if (!firstTime) firstTime = time(NULL);
  lastTime = time(NULL);
  nSent += l;
#endif
  for (int i = 0; i < l; i++) {
    if (emptyTrans) {
      emptyTrans = 0;
      OUT(base + TRANSMIT_HOLDING_REGISTER, s[i]);
    } else {
      outQ.data[outQ.head] = s[i];
      if (++outQ.head == O_QSIZE) outQ.head = 0;
      outQ.n++;
    }
  }
  return 0;
}

void portIclear() {
  cli();
  inQ.head = inQ.tail = 0;
  inQ.n = 0;
  sti();
  nPacketsWaiting = 0;
}

int portIcount() {
  return inQ.n;
}

int portOcount() {
  return outQ.n;
}

void __interrupt ISR_8250() {
  static int newPacket = 1, CRCbyte, esc;
  static int c;
#ifdef DEBUG
  nISRcalls++;
  if (outQ.n > maxOcount) maxOcount = outQ.n;
#endif
  for (;;) {
    switch (IN(base + INTERRUPT_ID_REGISTER) & 0x07) {
      case IIR_RX_DATA_READY_INTERRUPT:
	c = IN(base + RECEIVE_BUFFER_REGISTER);
        if (inQ.n == I_QSIZE) {
#ifdef DEBUG
          nIdropped++;
#endif
          continue;
        }
	inQ.data[inQ.head] = c;
	if (++inQ.head == I_QSIZE) inQ.head = 0;
	inQ.n++;
        if (newPacket) {
          newPacket = 0;
          CRCbyte = 1;
          break;
        }
        if (CRCbyte) {
          CRCbyte = 0;
          break;
        }
	if (esc) {
	  esc = 0;
	  if (c != FRAME_CHAR) {
            newPacket = 1;
            nPacketsWaiting++;
#ifdef DEBUG
            if (nPacketsWaiting > maxPacketsWaiting)
              maxPacketsWaiting = nPacketsWaiting;
#endif
          }
	} else if (c == FRAME_CHAR) esc = 1;
	break;
      case IIR_TX_HOLDING_REGISTER_INTERRUPT:
	if (!outQ.n) {
	  emptyTrans = 1;
	  break;
	}
	c = outQ.data[outQ.tail];
	if (++outQ.tail == O_QSIZE) outQ.tail = 0;
	OUT(base + TRANSMIT_HOLDING_REGISTER, c);
	outQ.n--;
	break;
      case IIR_LINE_STATUS_INTERRUPT:
	lineStatus = IN(base + LINE_STATUS_REGISTER);
	break;
      default:
#ifdef DEBUG
        if (inQ.n > maxIcount) maxIcount = inQ.n;
#endif
#ifdef IPXEMU
	if (nPacketsWaiting && nRecb && !IPXmutex) {
          IPXmutex++;
          OUT(0x20, 0x20);
	  sti();
          IPXcheck();
          IPXmutex--;
          return;
	}
#endif
	OUT(0x20, 0x20);
	return;
    }
  }
}

void __interrupt ISR_16550() {
  static int newPacket = 1, CRCbyte, esc;
  static int c, count;
#ifdef DEBUG
  nISRcalls++;
  if (outQ.n > maxOcount) maxOcount = outQ.n;
#endif
  for (;;) {
    switch (IN(base + INTERRUPT_ID_REGISTER) & 0x07) {
      case IIR_RX_DATA_READY_INTERRUPT:
	do {
	  c = IN(base + RECEIVE_BUFFER_REGISTER);
          if (inQ.n == I_QSIZE) {
#ifdef DEBUG
            nIdropped++;
#endif
            continue;
          }
	  inQ.data[inQ.head] = c;
	  if (++inQ.head == I_QSIZE) inQ.head = 0;
	  inQ.n++;
          if (newPacket) {
            newPacket = 0;
            CRCbyte = 1;
            continue;
          }
          if (CRCbyte) {
            CRCbyte = 0;
            continue;
          }
	  if (esc) {
	    esc = 0;
	    if (c != FRAME_CHAR) {
              newPacket = 1;
              nPacketsWaiting++;
#ifdef DEBUG
              if (nPacketsWaiting > maxPacketsWaiting)
                maxPacketsWaiting = nPacketsWaiting;
#endif
            }
	  } else if (c == FRAME_CHAR) esc = 1;
	} while (IN(base + LINE_STATUS_REGISTER) & LSR_DATA_READY);
	break;
      case IIR_TX_HOLDING_REGISTER_INTERRUPT:
	if (!outQ.n) {
	  emptyTrans = 1;
	  break;
	}
	for (count = 0; count < 15 && outQ.n; count++) {
	  c = outQ.data[outQ.tail];
	  if (++outQ.tail == O_QSIZE) outQ.tail = 0;
	  OUT(base + TRANSMIT_HOLDING_REGISTER, c);
          	  outQ.n--;
	}
	break;
      case IIR_LINE_STATUS_INTERRUPT:
	lineStatus = IN(base + LINE_STATUS_REGISTER);
	break;
      default:
#ifdef DEBUG
        if (inQ.n > maxIcount) maxIcount = inQ.n;
#endif
#ifdef IPXEMU
	if (nPacketsWaiting && nRecb && !IPXmutex) {
          IPXmutex++;
          OUT(0x20, 0x20);
	  sti();
          IPXcheck();
          IPXmutex--;
          return;
	}
#endif
	OUT(0x20, 0x20);
	return;
    }
  }
}
